import bpy

from ragdollc import registry

from ..vendor import bpx
from ..ui import window
from .. import commands
from . import with_ctrl_dialog


def get_one_transform_from_selection():
    selection = bpx.selection(type=(bpx.BpxBone, bpx.BpxObject))

    # Exclude any Ragdoll object
    selection = list(filter(
        lambda s: not s.type().startswith("rd"), selection)
    )

    if selection:
        return selection[-1]


def get_one_solver(context=None):
    context = context or bpy.context
    # Solver from current active/pinned object
    if context.object:
        xobj = bpx.BpxType(context.object)
        if bpx.is_type(xobj, "rdSolver"):
            return xobj

    # Solver from last selected marker
    markers = bpx.selection(type="rdMarker")
    if markers:
        marker = markers[-1]
        scene_ = registry.get("SceneComponent", marker.data["entity"])
        return bpx.alias(scene_.entity)

    # Last solver listed out from scene
    solvers = bpx.ls(type="rdSolver")
    if solvers:
        return solvers[-1]


class RetargetMenuOp(bpy.types.Operator):
    bl_idname = "ragdoll.ui_menu_retarget"
    bl_label = "Retarget"
    bl_options = {"INTERNAL", "UNDO"}
    bl_description = ("Set recording target to selected marker, "
                      "or hold Ctrl to open Retarget UI")

    @classmethod
    def poll(cls, context):
        if not bpx.ls(type="rdSolver"):
            cls.poll_message_set("No Ragdoll solvers in the current scene")
            return False
        return True

    @with_ctrl_dialog
    def invoke(self, context, event):
        bpy.ops.ragdoll.retarget_window()
        return {"FINISHED"}

    def execute(self, context):
        bpy.ops.ragdoll.retarget()  # See edit_marker.py:Retarget
        return {"FINISHED"}


class RetargetWindow(bpy.types.Operator):
    bl_idname = "ragdoll.retarget_window"
    bl_label = "Retarget Window"
    bl_options = {"REGISTER"}
    bl_description = "Show solver retargeting window"

    @classmethod
    def poll(cls, context):
        if not bpx.ls(type="rdSolver"):
            cls.poll_message_set("No Ragdoll solvers in the current scene")
            return False
        return True

    def execute(self, context):
        solver = get_one_solver(context)
        open_retarget_window(solver, context)
        return {"FINISHED"}


def open_retarget_window(solver: bpx.BpxType, context=None):
    context = context or bpy.context
    solver = solver.handle()
    solver_ui = solver.rdSolverUi
    solver_ui.targets_rebuild()

    prev_win = None
    if solver_ui.targets_window:
        for win in context.window_manager.windows:
            if str(hash(win)) == solver_ui.targets_window:
                prev_win = win
                break

    if prev_win:
        # Close previous retarget window if found, e.g. minimized.
        #
        _areas = prev_win.screen.areas
        _space = _areas[0].spaces[0]
        # Check ui status.
        # If it is being used to display something else, leave it be.
        _is_retarget_win = (
                len(_areas) == 1 and
                _areas[0].type == "PROPERTIES" and
                _space.context == "PHYSICS" and
                _space.pin_id == solver
        )
        if _is_retarget_win:
            with context.temp_override(window=prev_win):
                bpy.ops.wm.window_close()

    win = window.create_window(1200, 600, window.E_SPACE_PROPERTIES)
    solver_ui.targets_window = str(hash(win))

    win_area = win.screen.areas[0]
    win_space = win_area.spaces[0]
    win_space.show_region_header = False
    win_space.use_pin_id = True
    win_space.pin_id = solver
    win_space.context = "PHYSICS"


class Retarget(bpy.types.Operator):
    bl_idname = "ragdoll.ui_retarget"
    bl_label = "Set/Append Marker Target"
    bl_options = {"INTERNAL", "UNDO"}
    bl_description = "Set recording target to the selected marker"

    marker: bpy.props.IntProperty(
        options={"SKIP_SAVE"},
        description="Marker entity Id. If 0, a marker needs to be selected",
    )
    append: bpy.props.BoolProperty(
        default=False,
        options={"SKIP_SAVE"},
        description="Append recording target to marker",
    )

    @classmethod
    def poll(cls, context):
        if context.space_data.type == "PROPERTIES":
            # In Retarget UI.

            xobj = bpx.BpxType(context.object)
            if bpx.is_type(xobj, "rdSolver"):

                solver_ui = context.object.rdSolverUi
                if not len(solver_ui.targets):
                    cls.poll_message_set("Target list not built yet.")
                    return False

                if context.space_data.use_pin_id:
                    # Marker is set by UI, so just need to select transform.
                    if get_one_transform_from_selection():
                        return True

                    else:
                        cls.poll_message_set("Select one object or pose bone.")
                        return False
                else:
                    # If solver object is not pinned, it doesn't make sense
                    # to continue. Because the solver panel will no longer
                    # active after selecting a transform.
                    cls.poll_message_set("Please enable 'Pin ID'.")
                    return False
        else:
            # For Retarget UI only
            return False

    def execute(self, context):
        transform = get_one_transform_from_selection()
        marker = bpx.alias(self.marker)

        commands.retarget([transform], marker, append=self.append)

        self.report({"INFO"}, "Retargeted %s -> %s" % (marker, transform))
        return {"FINISHED"}


class Untarget(bpy.types.Operator):
    bl_idname = "ragdoll.ui_untarget"
    bl_label = "Remove Marker Target"
    bl_options = {"INTERNAL", "UNDO"}
    bl_description = "Remove all recording targets from selected marker"

    marker: bpy.props.IntProperty(
        options={"SKIP_SAVE"},
        description="Marker entity Id.",
    )

    @classmethod
    def poll(cls, context):
        if context.space_data.type == "PROPERTIES":
            # In Retarget UI.
            xobj = bpx.BpxType(context.object)
            if bpx.is_type(xobj, "rdSolver"):

                solver_ui = context.object.rdSolverUi
                if len(solver_ui.targets):
                    return True
                else:
                    cls.poll_message_set("Target list not built yet.")
                    return False
            else:
                cls.poll_message_set("Active object is not a solver.")
                return False
        else:
            # For Retarget UI only
            return False

    def execute(self, context):
        marker = bpx.alias(self.marker)
        commands.untarget(marker)
        return {"FINISHED"}


def install():
    bpy.utils.register_class(RetargetWindow)
    bpy.utils.register_class(RetargetMenuOp)
    bpy.utils.register_class(Retarget)
    bpy.utils.register_class(Untarget)


def uninstall():
    bpy.utils.unregister_class(RetargetWindow)
    bpy.utils.unregister_class(RetargetMenuOp)
    bpy.utils.unregister_class(Retarget)
    bpy.utils.unregister_class(Untarget)
